from time import sleep

import socketio

sio_follow = socketio.Client()


@sio_follow.event
def connect():
    pass


def add_job_to_socket(job_id: str):
    sio_follow.emit("add_job_to_socket", {"job_id": job_id})


def del_job_to_socket(job_id: str):
    sio_follow.emit("del_job_to_socket", {"job_id": job_id})


@sio_follow.on("to_client_follow")
def to_client_follow(data):
    print(data)


@sio_follow.on("to_client_follow_no_newline")
def to_client_follow_no_newline(data):
    print(data, end="")


@sio_follow.event
def disconnect():
    pass
    # print("disconnected from server")
    # TODO make an API call to tell that the server has disconnected


def sio_disconnect():
    if sio_follow.connected:
        sio_follow.disconnect()


def connect_to_server_socket(public_ip: str, port: str = "12346"):
    sio_disconnect()

    sleep(1)
    retries = 20
    while retries > 0:
        try:
            if sio_follow.connected:
                break
            sio_follow.connect(f"http://{public_ip}:{port}/", wait=True)
            break
        except Exception as e:
            print(f"{str(e)}")
            print("Connecting server unsuccessful, retry in 30s")
            sleep(5)
            retries -= 1

    if not retries:
        raise Exception(f"Failed to connect to server socket {public_ip}")



from colorama import Fore
from colorama import Style


def print_sending(msg: str):
    print(f"[SENDING]: {msg}")


def print_receiving(msg: str):
    print(f"[RECEIVING]: {msg}")


def print_launching(msg: str):
    print(f"[LAUNCHING]: {msg}")


def print_gearing(msg: str):
    print(f"[GEARING]: {msg}")


def print_success(msg: str):
    print(f"[SUCCESS]: {msg}")


def print_err(msg: str):
    print(Fore.RED + f"[ERROR]: {msg}")
    print(Style.RESET_ALL)


def print_warning(msg: str):
    print(Fore.YELLOW + f"[WARNING]: {msg}")
    print(Style.RESET_ALL)


def print_highlight(msg: str):
    print(Fore.GREEN + f"{msg}")
    print(Style.RESET_ALL)



from getpass import getpass
from typing import Optional
from typing import Tuple

from aibro.constant import IS_DEMO
from aibro.tools.prints import print_err
from aibro.tools.service.api_service import aibro_client
from aibro.tools.utils import sha256_encode

AUTHENTICATED_USER_ID = None


def check_authentication(access_token: str = None) -> Optional[Tuple[str, str]]:
    global AUTHENTICATED_USER_ID
    if not AUTHENTICATED_USER_ID:
        if IS_DEMO:
            email = input("Welcome! To launch a trial, please enter your email: ")
            data = {"email": email, "password": "demo"}
        elif access_token:
            data = {"access_token": access_token}
        else:
            email = input("Enter your email: ")
            password = getpass()
            encoded = sha256_encode(password)
            data = {"email": email, "password": encoded}
        try:
            resp = aibro_client.post_with_json_data(
                "v1/authenticate", data, handle_error=False
            )
            if resp.status_code == 402:
                AUTHENTICATED_USER_ID = resp.json()["user_id"]
                while True:
                    username = input("Create a username: ")
                    data = {
                        "user_id": AUTHENTICATED_USER_ID,
                        "username": username,
                    }
                    resp_2 = aibro_client.post_with_json_data(
                        "v1/create_username", data, handle_error=False
                    )
                    if resp_2.status_code != 200:
                        print(resp_2.text)
                        continue
                    print(resp_2.text)
                    break
            AUTHENTICATED_USER_ID = resp.json()["user_id"]
        except Exception as e:
            print_err(f"Authentication Error: {str(e)}")
            return None
    else:
        if IS_DEMO:
            print("Already entered email!")
        else:
            print("Already authenticated!")

    return AUTHENTICATED_USER_ID



import os

if os.path.exists(".env"):
    from dotenv import load_dotenv

    load_dotenv(".env")

SERVER_HOST = os.environ.get("SERVER_HOST", "localhost")

IS_LOCAL = os.environ.get("IS_LOCAL", "false") == "true"

IS_DEMO = os.environ.get("IS_DEMO", "false") == "true"

# TODO: move this limitation check to aibro server
INPUT_DATA_SIZE_LIMIT_MB = 1024 * 1.5

INF_HOST = os.environ.get("SERVER_HOST", "http://api.aipaca.ai")



from aibro.comm import Comm  # noqa
from aibro.inference import Inference  # noqa
from aibro.job import Job  # noqa
from aibro.training import Training  # noqa

# from aibro.model import Model  # noqa
# from aibro.training_data import Data  # noqa



import datetime
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
from tensorflow.keras.callbacks import History
from tensorflow.keras.models import load_model
from tensorflow.keras.models import Model as TF_Model

from aibro.comm import available_machines
from aibro.tools.auth import check_authentication
from aibro.tools.helpers.input_assertions import assert_batch_size_type
from aibro.tools.helpers.input_assertions import assert_data_type
from aibro.tools.helpers.input_assertions import assert_epochs_type
from aibro.tools.helpers.input_assertions import assert_is_tf_model
from aibro.tools.marketplace import _check_machines_exist
from aibro.tools.online_pickup import train_on_server
from aibro.tools.prints import print_err
from aibro.tools.prints import print_receiving
from aibro.tools.service.api_service import aibro_client
from aibro.tools.spot import _stop_instance
from aibro.tools.spot import save_cancel_job_reason
from aibro.tools.utils import datetime_to_timestamp_ms
from aibro.tools.utils import save_tensorboard_logs


class Training(object):
    @staticmethod
    def online_fit(
        model: TF_Model,
        train_X: np.ndarray,
        train_Y: np.ndarray,
        machine_ids: List[str] = None,
        batch_size: int = 1,
        epochs: int = 1,
        validation_data: Tuple[np.ndarray, np.ndarray] = None,
        description: str = "",
        cool_down_period_s: int = 0,
        fit_kwargs: Dict[str, Any] = {},
        directory_to_save_ckpt: str = None,
        directory_to_save_log: str = None,
        wait_request_s: int = 30,
        wait_new_job_create_s: int = -1,
        wait_new_job_create_interval: int = 10,
        record: bool = False,
        access_token=None,
    ):
        """[reference]
        https://doc.aipaca.ai/#online_fit
        """
        return online_fit(
            model,
            train_X,
            train_Y,
            machine_ids,
            batch_size,
            epochs,
            validation_data,
            description,
            cool_down_period_s,
            fit_kwargs,
            directory_to_save_ckpt,
            directory_to_save_log,
            wait_request_s,
            wait_new_job_create_s,
            wait_new_job_create_interval,
            record,
            access_token,
        )


def online_fit(
    model: TF_Model,
    train_X: np.ndarray,
    train_Y: np.ndarray,
    machine_ids: List[str] = None,
    batch_size: int = 1,
    epochs: int = 1,
    validation_data: Tuple[np.ndarray, np.ndarray] = None,
    description: str = "",
    cool_down_period_s: int = 0,
    fit_kwargs: Dict[str, Any] = {},
    directory_to_save_ckpt: str = None,
    directory_to_save_log: str = None,
    wait_request_s: int = 30,
    wait_new_job_create_s: int = -1,
    wait_new_job_create_interval: int = 10,
    record: bool = False,
    access_token: str = None,
) -> Tuple[Optional[str], Optional[TF_Model], Optional[History]]:
    job_id, trained_model, history = None, None, None
    try:
        # Check input types
        job_created_at_ms = datetime_to_timestamp_ms(datetime.datetime.now())
        train_X, train_Y, validation_data = _data_validation(
            model, train_X, train_Y, batch_size, epochs, validation_data
        )

        validation_X, validation_Y = None, None
        if validation_data:
            validation_X, validation_Y = validation_data

        job_id, history, model_file_content, tensorboard_logs, epoch = train_on_server(
            model=model,
            train_X=train_X,
            train_Y=train_Y,
            machine_ids=_check_machine_ids(machine_ids),
            batch_size=batch_size,
            epochs=epochs,
            validation_X=validation_X,
            validation_Y=validation_Y,
            description=description,
            fit_kwargs=fit_kwargs,
            cool_down_period=cool_down_period_s,
            job_created_at_ms=job_created_at_ms,
            wait_request_s=wait_request_s,
            wait_new_job_create_s=wait_new_job_create_s,
            wait_new_job_create_interval=wait_new_job_create_interval,
            directory_to_save_log=directory_to_save_log,
            record=record,
            access_token=access_token,
        )

        # if save ckpt not provided default not saving h5
        not_save_as_h5 = not directory_to_save_ckpt
        not_save_log = not directory_to_save_log
        # temperarily to save the model to '.' which will later get deleted
        # if not provided
        directory_to_save_ckpt = (
            directory_to_save_ckpt if directory_to_save_ckpt else "."
        )
        trained_model = _get_model_from_file(
            model_file_content=model_file_content,
            not_save_as_h5=not_save_as_h5,
            directory_to_save_ckpt=directory_to_save_ckpt,
            epoch=epoch,
        )
        save_tensorboard_logs(
            tensorboard_logs, not_save_log, directory_to_save_log, job_id
        )
        return job_id, trained_model, history

    except Exception as e:
        print_err(str(e))
        return job_id, trained_model, history


def _pick_machine_id():
    available_machines()
    machine_id = input("Enter machine_id: ")
    print(f"You have selected {machine_id}\n")
    return machine_id


def _check_machine_ids(machine_ids: List[str] = None) -> List[str]:
    while not machine_ids or len(machine_ids) == 0:
        print(
            "We detected that you didn't select a cloud machine. Please select one of the followings:"
        )
        machine_ids = [_pick_machine_id()]
    marketplace = aibro_client.get("v1/marketplace_machines").json()
    new_machine_ids: List[str] = [
        m for m in machine_ids if _check_machines_exist(m, marketplace)
    ]
    while len(new_machine_ids) == 0:
        print(
            "We detected that the cloud machine you selected is not available. \
                    Please select one of the followings:"
        )
        new_machine_ids = [_pick_machine_id()]
        if not _check_machines_exist(new_machine_ids[0], marketplace):
            new_machine_ids = []
    return new_machine_ids


def _get_model_from_file(
    model_file_content: Optional[bytes],
    not_save_as_h5: bool,
    directory_to_save_ckpt: Optional[str],
    epoch: int,
) -> Optional[TF_Model]:
    model = None
    if model_file_content:
        file_name = f"{directory_to_save_ckpt}/model_{epoch:04d}.h5"
        f = open(file_name, "wb")
        f.write(model_file_content)
        f.close()
        model = load_model(file_name)

        if not_save_as_h5:
            os.remove(file_name)
        else:
            print_receiving(f"Saved ckpt file to {file_name}")

    return model


def _data_validation(
    model: TF_Model,
    train_X: np.ndarray,
    train_Y: np.ndarray,
    batch_size: int = 1,
    epochs: int = 1,
    validation_data: Tuple[np.ndarray, np.ndarray] = None,
):
    assert_is_tf_model(model)
    train_X = assert_data_type(train_X, "train_X")
    train_Y = assert_data_type(train_Y, "train_Y")
    assert_batch_size_type(batch_size)
    assert_epochs_type(epochs)

    validation_X, validation_Y = None, None
    if validation_data is not None:
        validation_X, validation_Y = validation_data
        validation_X = assert_data_type(validation_X, "validation_X")
        validation_Y = assert_data_type(validation_Y, "validation_Y")
    return train_X, train_Y, (validation_X, validation_Y)


def cancel_job(job_id: str, e: str):
    check_authentication()
    try:
        print_err(f"Trying to cancel offline job: {job_id}")
        _stop_instance(job_id, e, endpoint="v1/training/stop_instance")
        save_cancel_job_reason(
            job_id,
            "User Called Cancel Job",
            endpoint="v1/training/save_cancel_job_reason",
        )
        print_err(f"Job {job_id} canceled successfully")
    except Exception as e:
        print_err(f"Canceling job failed: {str(e)}")


# def offline_fit(
#     model: TF_Model,
#     train_X: np.ndarray,
#     train_Y: np.ndarray,
#     machine_ids: List[str] = None,
#     batch_size: int = 1,
#     epochs: int = 1,
#     validation_data: Tuple[np.ndarray, np.ndarray] = None,
#     description="",
#     fit_kwargs: Dict[str, Any] = {},
#     cool_down_period_s: int = 0,
#     save_data=False,
#     data_name=None,
# ):
#     job_id = None
#     try:
#         # Check input types
#         _data_validation(model, train_X, train_Y, batch_size, epochs, validation_data)
#         validation_X, validation_Y = None, None
#         if validation_data:
#             validation_X, validation_Y = validation_data

#         job_id, _, _, _, _ = train_on_server(
#             model=model,
#             train_X=train_X,
#             train_Y=train_Y,
#             machine_ids=_check_machine_ids(machine_ids),
#             batch_size=batch_size,
#             epochs=epochs,
#             validation_X=validation_X,
#             validation_Y=validation_Y,
#             description=description,
#             fit_kwargs=fit_kwargs,
#             cool_down_period=cool_down_period_s,
#             fit_type="offline",
#         )
#         return job_id

#     except Exception as e:
#         print_err(str(e))
#         aibro_client.disconnect_socket()
#         return job_id



import os
import platform
import socket
import uuid
from typing import Any
from typing import Dict
from typing import Union

import requests
import urllib3.connection
from requests import Response

from aibro.constant import IS_LOCAL
from aibro.tools.sio import sio as sio_main
from aibro.tools.sio import sio_follow
from aibro.tools.sio import sio_follow_inf
from aibro.tools.sio import sio_spot
from aibro.tools.utils import ProgressUpload

# Set TCP keep alive options to avoid HTTP requests hanging issue
# Reference: https://stackoverflow.com/questions/61493298/flask-app-response-not-received-client-side
# Reference: https://stackoverflow.com/a/14855726/2360527
platform_name = platform.system()
orig_connect = urllib3.connection.HTTPConnection.connect


def patch_connect(self):
    orig_connect(self)
    if platform_name == "Linux" or platform_name == "Windows":
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 1),
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 3),
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5),
    elif platform_name == "Darwin":
        TCP_KEEPALIVE = 0x10
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        self.sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, 3)


urllib3.connection.HTTPConnection.connect = patch_connect


class APIClient:
    def __init__(self, host: str, port: int, sio):
        self.host = host
        self.port = port
        self.sio = sio

    def connect_to_socket(self):
        self.sio.connect_to_server_socket(self.host, self.port)

    def disconnect_socket(self):
        self.sio.sio_disconnect()

    def get(self, endpoint: str):
        try:
            resp = requests.get(f"http://{self.host}:{self.port}/{endpoint}")
        except Exception as e:
            return f"ERROR: {e}"
        if resp.status_code != 200:
            self._handle_error(resp)

        return resp

    def post_with_json_data(
        self, endpoint: str, json_data: Dict[str, Any], handle_error: bool = True
    ) -> Response:
        headers = {"Content-Type": "application/json"}
        resp = requests.post(
            f"http://{self.host}:{self.port}/{endpoint}",
            json=json_data,
            headers=headers,
        )
        if resp.status_code == 202:
            return resp
        if handle_error and resp.status_code != 200:
            self._handle_error(resp)
        return resp

    def post_with_raw_data(self, endpoint: str, raw_data: str) -> Response:
        headers = {"Content-Type": "application/json"}
        resp = requests.post(
            f"http://{self.host}:{self.port}/{endpoint}",
            data=raw_data,
            headers=headers,
        )
        if resp.status_code == 202:
            return resp
        if resp.status_code != 200:
            self._handle_error(resp)

        return resp

    def upload_file(self, upload_url: str, filepath: str):

        return requests.post(
            upload_url,
        )

    def post_with_byte_data(
        self,
        endpoint: str,
        json_data: Dict[str, Any] = None,
        filepath: str = None,
    ) -> Union[Response, str]:
        if not filepath:
            # prevent R/W from multiple processes
            random_postfix = str(uuid.uuid1())
            filepath = f"./user_data_{random_postfix}.json"
            with open(filepath, "w") as f:
                f.write(str(json_data))
        upload_url = f"http://{self.host}:{self.port}/{endpoint}"
        try:
            resp = requests.post(upload_url, data=ProgressUpload(filepath))
        except Exception as e:
            return f"ERROR: {e}"

        files = os.listdir(".")
        for file in files:
            if file.startswith("user_data_"):
                os.remove(os.path.join(".", file))
        if resp.status_code == 202:
            return resp
        if resp.status_code != 200:
            self._handle_error(resp)
        return resp

    def post_with_endpoint_only(self, endpoint: str) -> Response:
        resp = requests.post(
            f"http://{self.host}:{self.port}/{endpoint}",
        )
        if resp.status_code == 202:
            return resp

        if resp.status_code != 200:
            self._handle_error(resp)

        return resp

    def _handle_error(self, resp: Response):
        content = resp.content.decode("utf-8")
        self.sio.sio_disconnect()
        raise Exception(content)

    def add_job_to_socket(self, job_id: str):
        self.sio.add_job_to_socket(job_id)

    def del_job_to_socket(self, job_id: str):
        self.sio.del_job_to_socket(job_id)


class AIbroClient(APIClient):
    def __init__(self):
        host = (
            os.environ.get("SERVER_HOST", "localhost") if IS_LOCAL else "52.204.82.19"
        )
        port = 8000
        sio = sio_main
        super().__init__(host=host, port=port, sio=sio)

    def create_new_job(
        self, data: Dict[str, Any], endpoint="v1/inference/create_new_job"
    ) -> Response:
        headers = {"Content-Type": "application/json"}
        resp = requests.post(
            f"http://{self.host}:{self.port}/{endpoint}",
            json=data,
            headers=headers,
        )
        return resp

    def request_instance(
        self, data: Dict[str, Any], endpoint="v1/inference/request_instance"
    ) -> Response:
        return self.post_with_json_data(endpoint, data)

    def complete_inf_job(
        self, data: Dict[str, Any], endpoint="v1/complete_inf_job"
    ) -> Response:
        return self.post_with_json_data(endpoint, data)

    def update_clients(
        self, data: Dict[str, Any], endpoint="v1/update_clients"
    ) -> Response:
        return self.post_with_json_data(endpoint, data)

    def list_clients(
        self, data: Dict[str, Any], endpoint="v1/list_clients"
    ) -> Response:
        return self.post_with_json_data(endpoint, data)

    def spin_up_server(
        self, data: Dict[str, Any], endpoint="v1/inference/spin_up_server"
    ) -> Response:
        return self.post_with_json_data(endpoint, data)

    def get_serialization_details(self, data: Dict[str, Any]) -> Response:
        return self.post_with_json_data("v1/get_serialization_detail", data)


class SpotClient(APIClient):
    def __init__(self, is_follow_app=False, port=12345, sio=sio_spot):
        port = port
        sio = sio
        super().__init__(host="", port=port, sio=sio)

    def set_host(self, host: str):
        self.host = host


aibro_client = AIbroClient()
spot_client = SpotClient()
spot_client_follow = SpotClient(is_follow_app=True, port=12346, sio=sio_follow)
spot_client_follow_inf = SpotClient(is_follow_app=True, port=12348, sio=sio_follow_inf)
inf_ins_client = SpotClient(port=12347)



import os
import time
from time import sleep

import socketio

from aibro.tools.utils import printProgressBar

# logger=True, engineio_logger=True
sio_spot = socketio.Client()


@sio_spot.event
def connect():
    pass


def add_job_to_socket(job_id: str):
    sio_spot.emit("add_job_to_socket", {"job_id": job_id})


def del_job_to_socket(job_id: str):
    sio_spot.emit("del_job_to_socket", {"job_id": job_id})


@sio_spot.on("to_client_spot")
def to_client(data):
    print(data)


@sio_spot.event
def disconnect():
    pass
    # print("disconnected from server")
    # TODO make an API call to tell that the server has disconnected


@sio_spot.on("receive_model_return")
def receive_model_return(data):
    chunk = data["chunk"]
    json_mb = round(data["json_mb"], 2)
    job_id = data["job_id"]
    last_send = data["last_send"]
    progress = round(data["progress"], 2) if not last_send else json_mb
    time_start = data["time_start"]
    rate = round(progress / (time.time() - time_start), 2)
    unit = "MiB"
    filepath = f"model_and_data_unfinished_{job_id}.json"
    with open(filepath, "a") as f:
        f.write(chunk)
        printProgressBar(
            progress,
            json_mb,
            prefix="[RECEIVING]:",
            suffix=f"{progress}{unit}/{json_mb}{unit} [avg: {rate}{unit}/s]",
            length=10,
        )
    if last_send:
        os.rename(filepath, f"model_and_data_{job_id}.json")


def sio_disconnect():
    if sio_spot.connected:
        sio_spot.disconnect()


def connect_to_server_socket(public_ip: str, port: str = "12345"):
    sio_disconnect()

    sleep(1)
    retries = 20
    while retries > 0:
        try:
            if sio_spot.connected:
                break
            sio_spot.connect(f"http://{public_ip}:{port}/", wait=True)
            break
        except Exception as e:
            print(f"{str(e)}")
            print("Connecting server unsuccessful, retry in 30s")
            sleep(5)
            retries -= 1

    if not retries:
        raise Exception(f"Failed to connect to server socket {public_ip}")



import datetime
from time import sleep
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras.callbacks import History
from tensorflow.keras.losses import serialize as loss_serialize
from tensorflow.keras.metrics import serialize as metric_serialize
from tensorflow.keras.models import Model as TF_Model
from tensorflow.keras.optimizers import serialize as optimizer_serialize

from aibro.tools.auth import check_authentication
from aibro.tools.prints import print_err
from aibro.tools.prints import print_launching
from aibro.tools.prints import print_sending
from aibro.tools.service.api_service import aibro_client
from aibro.tools.spot import _stop_instance
from aibro.tools.spot import connect_to_spot_server_offline
from aibro.tools.spot import connect_to_spot_server_online
from aibro.tools.spot import save_cancel_job_reason
from aibro.tools.utils import datetime_to_timestamp_ms
from aibro.tools.utils import encode_data
from aibro.tools.utils import get_data_size_mb
from aibro.tools.utils import get_encoded_model_content


def train_on_server(  # noqa: C901
    batch_size: int,
    epochs: int,
    machine_ids: List[str],
    model: TF_Model = None,
    train_X=None,
    train_Y=None,
    validation_X=None,
    validation_Y=None,
    description="",
    fit_kwargs: Dict[str, Any] = {},
    cool_down_period: int = 0,
    fit_type: str = "online",
    job_created_at_ms: int = 0,
    wait_request_s: int = 30,
    wait_new_job_create_s: int = -1,
    wait_new_job_create_interval: int = 10,
    directory_to_save_log: str = None,
    record: bool = False,
    access_token: str = None,
) -> Tuple[Optional[str], Optional[History], Optional[bytes], Optional[str], int]:
    user_id = check_authentication(access_token)
    print("Please open https://aipaca.ai/training_jobs to track job status.")
    if not user_id:
        raise Exception()
    if record:
        print(
            "Turned on record mode. Your job data will be sent to the AIbro support team for debugging purposes."
        )
    history = None
    job_id = ""
    for i, machine_id in enumerate(machine_ids):
        try:
            last_try = i == len(machine_ids) - 1
            job_type = "training"
            data = {
                "user_id": user_id,
                "machine_id": machine_id,
                "description": description,
                "cool_down_period": cool_down_period,
                "fit_type": fit_type,
                "job_created_at_ms": job_created_at_ms,
                "wait_request_s": wait_request_s,
                "last_try": last_try,
                "job_type": job_type,
            }
            aibro_client.connect_to_socket()
            if i == 0:
                if wait_new_job_create_s <= 0:
                    wait_new_job_create_s = 99999999
                while wait_new_job_create_s:
                    resp = aibro_client.create_new_job(
                        data, endpoint="v1/training/create_new_job"
                    )
                    if resp.status_code == 403:
                        print_launching(
                            f"{resp.text} try again in another {wait_new_job_create_interval} seconds"
                        )
                        sleep(wait_new_job_create_interval)
                        wait_new_job_create_s -= wait_new_job_create_interval
                        continue
                    break
                if resp.status_code != 200:
                    raise Exception(resp.text)
                job_id = resp.json()["job_id"]
                print_launching(f"Starting training job: {job_id}")
            print_launching(f"Requesting {machine_id} to be ready...")
            data["job_id"] = job_id
            aibro_client.add_job_to_socket(data["job_id"])
            resp = aibro_client.request_instance(
                data, endpoint="v1/training/request_instance"
            )
            aibro_client.del_job_to_socket(data["job_id"])
            aibro_client.disconnect_socket()
            if resp.status_code != 200:
                if i == len(machine_ids) - 1:
                    raise Exception(resp.text)
                continue
            server_public_ip = resp.json()["server_public_ip"]
            was_idle = resp.json()["was_idle"]
            if was_idle:
                print(f"Found an cooling {machine_id}, ready to train now 🎉\n")
            else:
                data_spin_up = {
                    "job_id": job_id,
                }
                resp = aibro_client.spin_up_server(
                    data_spin_up, endpoint="v1/training/spin_up_server"
                )
                print(f"Your {machine_id} is now ready 🎉\n")

            print_sending("Serializing your model...")
            loss = model.loss  # type:ignore
            optimizer = model.optimizer  # type:ignore

            loss_config = loss_serialize(loss) if not isinstance(loss, str) else loss
            optimizer_config = (
                optimizer_serialize(optimizer)
                if not isinstance(optimizer, str)
                else optimizer
            )

            metrics = vars(model.compiled_metrics)["_metrics"]  # type:ignore
            metrics = list(np.array(metrics).flatten())
            metrics_config = [
                metric_serialize(metric) if not isinstance(metric, str) else metric
                for metric in metrics
            ]

            if model:
                encoded_model_file_content = get_encoded_model_content(model)

            train_X_shape = list(train_X.shape)
            train_Y_shape = list(train_Y.shape)
            valid_X_shape = None
            valid_Y_shape = None
            if validation_X is not None:
                valid_X_shape = list(validation_X.shape)
                valid_Y_shape = list(validation_Y.shape)

            print_sending("Serializing your data...")
            (
                encoded_train_X,
                encoded_train_Y,
                encoded_validation_X,
                encoded_validation_Y,
            ) = encode_data(train_X, train_Y, validation_X, validation_Y)
            m_d_serialized_at_ms = datetime_to_timestamp_ms(datetime.datetime.now())

            data = {
                "tensorflow_version": tf.__version__,
                "encoded_model_content": encoded_model_file_content if model else None,
                "encoded_train_X": encoded_train_X,
                "encoded_train_Y": encoded_train_Y,
                "encoded_validation_X": encoded_validation_X,
                "encoded_validation_Y": encoded_validation_Y,
                "train_X_shape": train_X_shape,
                "train_Y_shape": train_Y_shape,
                "valid_X_shape": valid_X_shape,
                "valid_Y_shape": valid_Y_shape,
                "user_id": user_id,
                "optimizer_config": optimizer_config,
                "loss_config": loss_config,
                "metrics_config": metrics_config,
                "batch_size": batch_size,
                "epochs": epochs,
                "fit_kwargs": fit_kwargs,
                "job_id": job_id,
                "m_d_serialized_at_ms": m_d_serialized_at_ms,
                "record": record,
            }
            print_sending("Sending model and data to the server to start training")

            if fit_type == "online":
                (
                    history,
                    model_file_content,
                    tensorboard_logs,
                    epoch,
                ) = connect_to_spot_server_online(
                    data=data,
                    endpoint="v1/model_and_dataset",
                    job_id=job_id,
                    server_public_ip=server_public_ip,
                    directory_to_save_log=directory_to_save_log,
                )
                results_received_at_ms = datetime_to_timestamp_ms(
                    datetime.datetime.now()
                )
                post_train_detail = {
                    "job_id": job_id,
                    "size_m_d_trans_mb": get_data_size_mb(data),
                    "results_received_at_ms": results_received_at_ms,
                }
                aibro_client.connect_to_socket()
                aibro_client.get_serialization_details(post_train_detail)
                aibro_client.disconnect_socket()

                return job_id, history, model_file_content, tensorboard_logs, epoch
            else:
                connect_to_spot_server_offline(
                    data,
                    endpoint="v1/model_and_dataset",
                    job_id=job_id,
                    server_public_ip=server_public_ip,
                )
                return job_id, None, None, None, -1

        except Exception as e:
            if "No enough" in str(e) and i < len(machine_ids) - 1:
                reason = (
                    f"machine_id: {machine_id} not available in AWS. Try the next one"
                )
                print_launching(reason)
                aibro_client.disconnect_socket()
                continue
            else:
                reason = f"{str(e)}"
                print_err(reason)
                if job_id:
                    save_cancel_job_reason(
                        job_id, reason, endpoint="v1/training/save_cancel_job_reason"
                    )
            aibro_client.disconnect_socket()
            return job_id, history, None, None, -1
        except KeyboardInterrupt:
            if job_id:
                print_err(f"Trying to cancelled online job: {job_id}")
                _stop_instance(
                    job_id,
                    "User KeyboardInterrupt",
                    endpoint="v1/training/stop_instance",
                )
                save_cancel_job_reason(
                    job_id,
                    "User KeyboardInterrupt",
                    endpoint="v1/training/save_cancel_job_reason",
                )
                print_err(f"Job {job_id} successfully canceled")
            else:
                print_err("Job cancelled before server is ready")
            return job_id, None, None, None, -1
    return job_id, None, None, None, -1



import copy
import datetime
import os
import shutil
import sys
import threading
from pathlib import Path
from time import sleep
from typing import List
from typing import Union

from requests.models import Response

from aibro.constant import INF_HOST
from aibro.tools.auth import check_authentication
from aibro.tools.marketplace import _check_machines_exist
from aibro.tools.prints import print_err
from aibro.tools.prints import print_highlight
from aibro.tools.prints import print_launching
from aibro.tools.prints import print_sending
from aibro.tools.prints import print_warning
from aibro.tools.service.api_service import aibro_client
from aibro.tools.service.api_service import inf_ins_client
from aibro.tools.service.api_service import spot_client_follow_inf
from aibro.tools.service.api_service import SpotClient
from aibro.tools.spot import _stop_instance
from aibro.tools.spot import save_cancel_job_reason
from aibro.tools.utils import datetime_to_timestamp_ms
from aibro.tools.utils import subprocess_to_stdout
from aibro.tools.utils import zipdir


class Inference(object):
    def __init__(
        self,
        model_name: str,
        machine_id_config: Union[str, dict],
    ):
        self.model_name = model_name
        self.machine_id_config = machine_id_config

    @staticmethod
    def deploy(
        artifacts_path: Union[str, list],
        model_name: str = None,
        machine_id_config: Union[str, dict] = None,
        dryrun: bool = False,
        cool_down_period_s: int = 600,
        client_ids: List[str] = [],
        access_token: str = None,
        description: str = "",
        wait_request_s: int = 30,
    ) -> str:
        if dryrun:
            if isinstance(artifacts_path, list):
                artifacts_path = _pack_inf_repo(artifacts_path, zip=False)
            assert isinstance(
                artifacts_path, str
            ), "artifacts_path must be either str or list"
            return _dryrun_helper(artifacts_path)
        try:
            assert model_name, "model_name does not exist"
            assert machine_id_config, "machine_id_config does not exist"
            if isinstance(client_ids, str):
                client_ids = [client_ids]
            job_created_at_ms = datetime_to_timestamp_ms(datetime.datetime.now())
            _validate_inference_machine_ids(machine_id_config)
            user_id = check_authentication(access_token)
            print("Please open https://aipaca.ai/inference_jobs to track job status.")
            if not user_id:
                raise Exception("Authentication failed")

            job_type = "inference"
            data = {
                "user_id": user_id,
                "machine_id_config": machine_id_config,
                "model_name": model_name,
                "description": description,
                "job_created_at_ms": job_created_at_ms,
                "job_type": job_type,
                "wait_request_s": wait_request_s,
                "cool_down_period_s": cool_down_period_s,
                "client_ids": client_ids,
            }
            job_id, machine_id_config = _create_new_inf_job(data)
            data["job_id"] = job_id
            data["machine_id_config"] = machine_id_config
            server_public_ips = _request_instance(data)
            data["server_public_ips"] = server_public_ips
            _launch_inf_server(data)
            print_sending("Serializing your artifacts...")
            zip_path = _pack_inf_repo(artifacts_path)
            username, access_token = _deploy_inf_model(
                server_public_ips, job_id, user_id, zip_path, model_name
            )
            api_url = _create_api_url(username, client_ids, model_name)
            client_str = ""
            public = len(client_ids) == 0
            if not public:
                client_str = " for one of clients"
            print_highlight(f"Your Inference API URL{client_str}: {api_url}")
            return api_url
        except KeyboardInterrupt:
            error = "User KeyboardInterrupt"
            job_id = None if "job_id" not in locals() else job_id
            server_public_ips = (
                None if "server_public_ips" not in locals() else server_public_ips
            )
            _handle_exception(error, job_id, server_public_ips)
            return error
        except Exception as e:
            error = str(e)
            job_id = None if "job_id" not in locals() else job_id
            server_public_ips = (
                None if "server_public_ips" not in locals() else server_public_ips
            )
            _handle_exception(error, job_id, server_public_ips)
            return error

    @staticmethod
    def complete(
        model_name: str = None,
        job_id: str = None,
        access_token: str = None,
    ):
        assert (
            job_id or model_name
        ), "To complete a inference job, you should input either its model_name or job_id"
        user_id = check_authentication(access_token)
        data = {"user_id": user_id, "model_name": model_name, "job_id": job_id}
        resp = aibro_client.complete_inf_job(data)
        if resp.status_code != 200:
            print_err(resp.text)
        model_name = resp.json()["model_name"]
        job_id = resp.json()["job_id"]
        print(
            f"Inference job {job_id}, with model {model_name}, successfully completed."
        )

    @staticmethod
    def update_clients(
        add_client_ids: Union[str, List[str]] = [],
        remove_client_ids: Union[str, List[str]] = [],
        be_public: bool = False,
        model_name: str = None,
        job_id: str = None,
        access_token: str = None,
    ) -> List[str]:
        assert (
            job_id or model_name
        ), "To find a inference job, you should input either its model_name or job_id"
        user_id = check_authentication(access_token)
        if isinstance(add_client_ids, str):
            add_client_ids = [add_client_ids]
        if isinstance(remove_client_ids, str):
            remove_client_ids = [remove_client_ids]
        data = {
            "user_id": user_id,
            "model_name": model_name,
            "job_id": job_id,
            "add_client_ids": add_client_ids,
            "remove_client_ids": remove_client_ids,
            "be_public": be_public,
        }
        resp = aibro_client.update_clients(data)
        if resp.status_code != 200:
            error = str(resp.text)
            print_err(error)
            raise Exception(error)
        print("Update client succeeded!")
        client_ids = resp.json()["client_ids"]
        print_highlight(f"Current client ids: {client_ids}")
        return client_ids

    @staticmethod
    def list_clients(
        model_name: str = None,
        job_id: str = None,
        access_token: str = None,
    ) -> List[str]:
        assert (
            job_id or model_name
        ), "To find a inference job, you should input either its model_name or job_id"
        user_id = check_authentication(access_token)
        data = {
            "user_id": user_id,
            "model_name": model_name,
            "job_id": job_id,
        }
        resp = aibro_client.list_clients(data)
        if resp.status_code != 200:
            error = str(resp.text)
            raise Exception(error)
        client_ids = resp.json()["client_ids"]
        print_highlight(f"Current client ids: {client_ids}")
        return client_ids


def _detect(repo_path: str):  # if false,raise err_msg, else return data file's name
    FILE_LIST = ["predict.py", "model", "requirements.txt"]
    for file in FILE_LIST:
        file_path = os.path.join(repo_path, file)
        if not os.path.exists(file_path):
            return False
    for file in os.listdir(repo_path):  # data folder/file is optional
        if file[:4] == "data":
            return True
    return True


def _dryrun_helper(
    artifacts_path: str,
):  # previous parameter is Union[str, list], noqa
    """
    Check if the artifacts_path given by user is valid。
    Test: check if predict.py can return inference result
    """
    err_msg = "Invalid repo structure (checkout https://doc.aipaca.ai for details)."
    check = _detect(artifacts_path)  # firt, check if the repo structure is our format
    aibro_dir_path = os.path.split(os.path.realpath(__file__))[0]
    inf = os.path.join(aibro_dir_path, "scripts", "infer_command.sh")
    call = os.path.join(aibro_dir_path, "scripts", "call_predict.sh")
    artifacts_path_abspath = os.path.abspath(artifacts_path)
    if check:  # if it is suitable
        try:
            bash = f"bash {inf} {artifacts_path_abspath}"
            subprocess_to_stdout(
                [bash],
                raise_error_if_any=True,
                line_to_stop_streaming="Dependencies installed!",
            )
            bash = f"bash {call} {artifacts_path_abspath}"
            subprocess_to_stdout(
                [bash],
                raise_error_if_any=True,
                line_to_stop_streaming="Prediction finished",
            )
        except Exception as e:
            raise Exception(f"DRYRUN TEST: failed - {str(e)}")
    else:
        raise Exception(f"DRYRUN TEST: failed - {err_msg}")

    print_highlight("DRYRUN TEST: passed")


def _create_new_inf_job(data):
    aibro_client.connect_to_socket()
    resp = aibro_client.create_new_job(data)
    if resp.status_code != 200:
        raise Exception(resp.text)
    job_id = resp.json()["job_id"]
    contain_spot = False
    if isinstance(data["machine_id_config"], str):
        if ".od" not in data["machine_id_config"]:
            contain_spot = True
    elif isinstance(data["machine_id_config"], dict):
        for k in data["machine_id_config"]:
            if ".od" not in data["machine_id_config"][k]:
                contain_spot = True
    if contain_spot:
        print_warning(
            f"{data['machine_id_config']} contained spot instances. Used its on-demand type instead"
        )
    machine_id_config = resp.json()["machine_id_config"]
    return job_id, machine_id_config


def _request_instance(data):
    machine_id_config = data["machine_id_config"]
    job_id = data["job_id"]
    belong = "public" if len(data["client_ids"]) == 0 else "private"
    print_launching(f"Starting {belong} inference job: {job_id}")
    print_launching(f"Requesting {str(machine_id_config)} to be ready...")
    data["job_id"] = job_id
    aibro_client.add_job_to_socket(data["job_id"])
    resp = aibro_client.request_instance(data)
    aibro_client.del_job_to_socket(data["job_id"])
    aibro_client.disconnect_socket()
    if resp.status_code != 200:
        raise Exception(resp.text)
    server_public_ips = resp.json()["server_public_ips"]
    return server_public_ips


def _launch_inf_server(data):
    # TODO: handle exception happened here
    machine_id_config = data["machine_id_config"]
    job_id = data["job_id"]
    data_spin_up = {
        "job_id": job_id,
    }
    aibro_client.spin_up_server(data_spin_up)
    print(f"Your {machine_id_config} instances are now ready 🎉\n")


def _pack_inf_repo(artifacts_path, zip=True):
    dir = "."
    if isinstance(artifacts_path, str):
        artifacts_dir_path = copy.deepcopy(artifacts_path)
        artifacts_path = [
            os.path.join(artifacts_dir_path, f) for f in os.listdir(artifacts_dir_path)
        ]
    # user passed a list of files
    artifacts_path_new = os.path.join(dir, "aibro_repo")
    counter = 0
    while os.path.exists(artifacts_path_new):
        # if aibro repo already exist in the current root, create a folder copy and create new aibro_repo in there
        counter += 1
        copy_dir_path = os.path.join(dir, f"aibro_repo_copy{counter}")
        artifacts_path_new = os.path.join(copy_dir_path, "aibro_repo")
        if not os.path.exists(copy_dir_path):
            os.mkdir(copy_dir_path)
    if not os.path.exists(artifacts_path_new):
        os.mkdir(artifacts_path_new)
    for f in artifacts_path:
        if os.path.isdir(f):
            folder_name = Path(f).name
            folder_path = os.path.join(artifacts_path_new, folder_name)
            shutil.copytree(f, folder_path)
        else:
            shutil.copy(f, artifacts_path_new)
    artifacts_path = artifacts_path_new
    if not zip:
        return artifacts_path
    zip_name = "aibro_repo.zip"
    zip_path = os.path.join(dir, zip_name)
    if os.path.exists(zip_path):
        os.remove(zip_path)
    zipdir(artifacts_path, zip_path)
    return zip_path


def _deploy_inf_model(server_public_ips, job_id, user_id, zip_path, model_name):
    threads = []
    for server_public_ip in server_public_ips:
        t = threading.Thread(
            target=_deploy_inf_model_setup_helper,
            args=(server_public_ip, job_id, user_id, zip_path),
        )
        t.start()
        threads.append(t)
        sleep(5)
    for t in threads:
        t.join()
    resp_1 = aibro_client.post_with_json_data(
        "v1/get_userinfo", json_data={"user_id": user_id}
    )
    resp_2 = aibro_client.post_with_json_data(
        "v1/add_active_model",
        json_data={"user_id": user_id, "job_id": job_id, "model_name": model_name},
    )
    if resp_1.status_code != 200 or resp_2.status_code != 200:
        reason = resp_1.text if resp_1.status_code != 200 else resp_2.text
        return _terminate_inference_job(job_id, reason, server_public_ips)
    username = resp_1.json()["username"]
    access_token = resp_1.json()["access_token"]
    return username, access_token


def _deploy_inf_model_setup_helper(
    server_public_ip: str,
    job_id: str,
    user_id: str,
    zip_path: str,
):
    try:
        # setup inference
        spot_client_follow_inf_copy = SpotClient(port=12347)
        spot_client_follow_inf_copy.set_host(server_public_ip)
        spot_client_follow_inf_copy.connect_to_socket()
        resp = spot_client_follow_inf_copy.post_with_byte_data(
            f"v1/{user_id}/inf_setup", filepath=zip_path
        )
        if isinstance(resp, Response) and resp.status_code != 200:
            return _terminate_inference_job(job_id, resp.text, [server_public_ip])
        # disconnect everything
        spot_client_follow_inf_copy.disconnect_socket()
        # _disconnect_inf_follow_app(job_id, follow_server_thread)
    except Exception as e:
        error = str(e)
        _stop_instance(job_id, error, server_public_ip)
        save_cancel_job_reason(job_id, error)
        print_err(f"Job {job_id} successfully canceled")
        return error


def _create_api_url(username: str, client_ids: list, model_name: str):
    client_id = "public" if len(client_ids) == 0 else client_ids[0]
    port_str = ":8000" if "http" not in INF_HOST else ""
    api_url = f"{INF_HOST}{port_str}/v1/{username}/{client_id}/{model_name}/predict"
    return api_url


def _terminate_inference_job(job_id, reason, server_public_ips: list = []):
    print_err(reason)
    save_cancel_job_reason(job_id, reason)
    for server_ip in server_public_ips:
        resp_stop_spot = _stop_instance(job_id, reason, server_ip)
        if resp_stop_spot.status_code != 200:
            print_err(f"Failed to terminate inference server: {resp_stop_spot.text}")
    inf_ins_client.disconnect_socket()
    raise Exception(reason)


def _validate_inference_machine_ids(machine_id_config: Union[str, dict]) -> bool:
    if isinstance(machine_id_config, str):
        machine_ids = [machine_id_config]
    elif isinstance(machine_id_config, dict):
        machine_ids = list(machine_id_config.values())
        assert len(machine_ids) > 0, "Please input at least one machine id"
    else:
        err_msg = "We detected that the cloud machine you selected is not available for inference job"
        raise Exception(err_msg)
    # TODO: check inference machine_ids from markertplace
    marketplace = aibro_client.get("v1/marketplace_machines").json()
    for m_id in machine_ids:
        assert _check_machines_exist(
            m_id, marketplace
        ), f"machine {m_id} not exist or lack of capacity"
    return True


def _handle_exception(error, job_id, server_public_ips):
    if job_id:
        print_err(f"Trying to cancel inference job: {job_id}")
        if not server_public_ips or len(server_public_ips) == 0:
            _stop_instance(job_id, error)
        else:
            for server_public_ip in server_public_ips:
                _stop_instance(job_id, error, server_public_ip)
        save_cancel_job_reason(job_id, error)
        print_err(f"Job {job_id} successfully canceled")
    else:
        print_err(error)


def _disconnect_inf_follow_app(job_id: str, follow_server_thread: threading.Thread):
    follow_server_thread.join(timeout=1)
    if follow_server_thread.is_alive():
        print("[Warning]: follow server didn't join succussfully")
    try:
        spot_client_follow_inf.del_job_to_socket(job_id)
        spot_client_follow_inf.disconnect_socket()
    except Exception as e:
        if "google.colab" not in sys.modules:
            print(f"[WARNING]: spot_client_follow: {e}")



import json
import os
import sys
import threading
from time import sleep
from typing import Any
from typing import Dict
from typing import Optional

from requests.models import Response

from aibro.tools.service.api_service import aibro_client
from aibro.tools.service.api_service import spot_client
from aibro.tools.service.api_service import spot_client_follow
from aibro.tools.service.api_service import SpotClient
from aibro.tools.utils import unpack_base64_encoded_pickle


def _remove_old_m_d_if_exist(filename):
    try:
        os.remove(filename)
        files = os.listdir(".")
        for file in files:
            if file.startswith("model_and_data_"):
                os.remove(os.path.join(".", file))
    except Exception:
        pass


def _find_filename(job_id):
    return f"./model_and_data_{job_id}.json"


def try_get_model_json(file_name, retries=-1, interval=5):
    json_str = ""
    retries = retries if retries > 0 else 9999999
    while retries:
        try:
            f = open(file_name)
            json_str = f.read()
            counter = -1
            while json_str[counter] == "\n":
                counter -= 1
            assert json_str[counter] == "}"
            break
        except Exception:
            sleep(interval)
            retries -= 1
            pass
    if json_str == "":
        raise Exception(
            "Failed to receive trained model data. This issue could be caused by slow internet."
        )
    model_and_data_json = json.loads(json_str)
    return model_and_data_json


def connect_to_spot_server_online(
    data: Dict[str, Any],
    endpoint: str,
    job_id: str,
    server_public_ip: str,
    directory_to_save_log: Optional[str],
):

    spot_client_follow.set_host(server_public_ip)
    spot_client_follow.connect_to_socket()
    spot_client_follow.add_job_to_socket(job_id)
    spot_client.set_host(server_public_ip)

    filename = _find_filename(job_id)
    _remove_old_m_d_if_exist(filename)

    follow_server_thread = threading.Thread(
        target=spot_client_follow.post_with_json_data,
        args=("v1/follow", {"job_id": job_id}),
    )
    follow_server_thread.start()

    # start training
    spot_client.post_with_byte_data(endpoint, data)
    # download dumped data
    spot_client.connect_to_socket()
    spot_client.add_job_to_socket(job_id)
    download_tensorflow_log = directory_to_save_log
    spot_client.post_with_json_data(
        "v1/download_train_result",
        {
            "job_id": job_id,
            "user_id": data["user_id"],
            "download_tensorflow_log": download_tensorflow_log,
        },
    )

    # data will be returned by socket
    model_and_data_json = try_get_model_json(filename)
    # decode data
    model_history = model_and_data_json["model_history"]
    history = unpack_base64_encoded_pickle(model_history)
    model_file_content = unpack_base64_encoded_pickle(
        model_and_data_json["model_file_content"]
    )
    tensorboard_logs = (
        model_and_data_json["tensorboard_logs"]
        if "tensorboard_logs" in model_and_data_json
        else None
    )
    epoch = model_and_data_json["epoch"]
    # got everything and disconnect socket
    _disconnect_everything(
        follow_server_thread, spot_client, spot_client_follow, job_id
    )
    return (
        history,
        model_file_content,
        tensorboard_logs,
        epoch,
    )


def connect_to_spot_server_offline(
    data: Dict[str, Any], endpoint: str, job_id: str, server_public_ip: str
):
    spot_client.set_host(server_public_ip)
    spot_client.connect_to_socket()
    # resp = spot_client.post_with_json_data(endpoint, data)
    resp = spot_client.post_with_byte_data(endpoint, data)
    if isinstance(resp, Response) and resp.status_code == 202:
        print(resp.content.decode("utf-8"))
    spot_client.disconnect_socket()


def _stop_instance(
    job_id: str, e: str, instance_ip: str = None, endpoint="v1/inference/stop_instance"
):
    data = {"job_id": job_id, "error": e, "instance_ip": instance_ip}
    aibro_client.connect_to_socket()
    resp_stop_spot = aibro_client.post_with_json_data(endpoint, data)
    aibro_client.disconnect_socket()
    return resp_stop_spot


def save_cancel_job_reason(
    job_id: str, reason: str, endpoint="v1/inference/save_cancel_job_reason"
):
    data = {"job_id": job_id, "reason": reason}
    aibro_client.post_with_json_data(endpoint, data)


def _disconnect_everything(
    follow_server_thread: threading.Thread,
    spot_client: SpotClient,
    spot_client_follow: SpotClient,
    job_id: str,
):
    follow_server_thread.join(timeout=1)
    if follow_server_thread.is_alive():
        print("[Warning]: follow server didn't join succussfully")
    try:
        spot_client.del_job_to_socket(job_id)
        spot_client.disconnect_socket()
    except Exception as e:
        if "google.colab" not in sys.modules:
            print(f"[WARNING]: spot_client: {e}")
    try:
        spot_client_follow.del_job_to_socket(job_id)
        spot_client_follow.disconnect_socket()
    except Exception as e:
        if "google.colab" not in sys.modules:
            print(f"[WARNING]: spot_client_follow: {e}")



from aibro.tools.auth import check_authentication
from aibro.tools.service.api_service import AIbroClient


aibro_client = AIbroClient()


class Comm(object):
    @staticmethod
    def send_message(email: str, feedback_message: str, category: str = "random"):
        """[reference]
        https://doc.aipaca.ai/#send_message
        """
        return send_message(email, feedback_message, category)

    @staticmethod
    def available_machines():
        """[reference]
        https://doc.aipaca.ai/#available_machines
        """
        return available_machines()


def send_message(email: str, feedback_message: str, category: str = "random"):
    if category not in set(["random", "feature_request", "bug report"]):
        raise Exception(
            "Message category has to be one of the ['random', 'feature_request', 'bug report']"
        )
    user_id = check_authentication()
    data = {
        "email": email,
        "user_id": user_id,
        "message": feedback_message,
        "category": category,
    }
    aibro_client.post_with_json_data("v1/feedback_email", data)
    print("Thanks for your feedback!")


def available_machines():
    resp = aibro_client.get("v1/marketplace_machines")
    print("Available Resources:")
    for name, spec in resp.json().items():
        limit = spec["CAPACITY"]
        if limit <= 0:
            continue

        GPU_type = spec["GPU"]
        num_vCPU = spec["CPU CORES"]
        pricing = spec["PRICING"]
        availability = spec["AVAILABILITY"]
        print(
            f"{f'Machine Id: {name}' : <26} {f'GPU Type: {GPU_type}' : <18} {f'num_vCPU: {num_vCPU}' : <15} "
            f"{f'cost: {pricing}' : <15} {f'capacity: {limit}' : <10} {f'availability: {availability * 100}%' : <15}"
            # noqa
        )



from aibro.constant import INPUT_DATA_SIZE_LIMIT_MB
from aibro.tools.utils import get_data_size_mb


def assert_is_tf_model(model):
    assert model.__class__.__module__.startswith(
        "tensorflow"
    ) or model.__class__.__module__.startswith(
        "keras"
    ), "ERROR: model must be a tf/keras model"


def assert_data_type(data, data_name):
    if type(data).__module__ == "numpy":
        size = get_data_size_mb(data)
        assert (
            size <= INPUT_DATA_SIZE_LIMIT_MB
        ), f"{data_name} has size {size} MB exceeding maximum limit of 1G"
    else:
        raise Exception(f"{data_name} must be numpy type")
    return data


def assert_batch_size_type(batch_size):
    assert type(batch_size) == int, "Incorrect type for batch_size (int)"


def assert_epochs_type(epochs):
    assert type(epochs) == int, "Incorrect type for epochs (int)"



import fnmatch
import os

from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_py import build_py as build_py_orig


excluded = [
    "aibro/data_operation.py",
    "aibro/model.py",
    "aibro/offline_pickup.py",
    "aibro/training_data.py",
    "tests/*",
]


class build_py(build_py_orig):
    """
    Used to exclude specific files
    Reference: https://stackoverflow.com/questions/35115892/how-to-ex\
        clude-a-single-file-from-package-with-setuptools-and-setup-py
    """

    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        return [
            (pkg, mod, file)
            for (pkg, mod, file) in modules
            if not any(fnmatch.fnmatchcase(file, pat=pattern) for pattern in excluded)
        ]


with open("PyPiReadMe.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_requirement_txt():
    thelibFolder = os.path.dirname(os.path.realpath(__file__))
    requirementPath = thelibFolder + "/requirements_deploy.txt"
    if os.path.isfile(requirementPath):
        with open(requirementPath) as f:
            install_requires = f.read().splitlines()
    return install_requires


setup(
    name="aibro",
    version="1.1.0",
    author="AIpaca.ai",
    author_email="hello@aipaca.ai",
    description="Serverless model training",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AIpaca-Inc/AIbro_lib",
    project_urls={
        "Bug Tracker": "https://github.com/AIpaca-Inc/AIbro_lib/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=("tests")),
    package_data={"": ["*.sh"]},
    cmdclass={"build_py": build_py},
    install_requires=get_requirement_txt(),
    scripts=["aibro/scripts/call_predict.sh", "aibro/scripts/infer_command.sh"],
    python_requires=">=3.6",
)



import base64
import datetime
import hashlib
import math
import os
import pickle
import shutil
import subprocess
import sys
import time
import uuid
import zipfile
import zlib
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Union

from objsize import get_deep_size
from tensorflow.keras.losses import deserialize as loss_deserialize
from tensorflow.keras.metrics import deserialize as metric_deserialize
from tensorflow.keras.models import load_model
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import deserialize as optimizer_deserialize

from aibro.tools.prints import print_receiving

if os.name == "nt":
    import ctypes  # noqa: F401
    import msvcrt  # noqa: F401

    class _CursorInfo(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]


def hide_cursor():
    if os.name == "nt":
        ci = _CursorInfo()
        handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
        ci.visible = False
        ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
    elif os.name == "posix":
        sys.stdout.write("\033[?25l")
        sys.stdout.flush()


def show_cursor():
    if os.name == "nt":
        ci = _CursorInfo()
        handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
        ci.visible = True
        ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
    elif os.name == "posix":
        sys.stdout.write("\033[?25h")
        sys.stdout.flush()


class ModelAndDataset(NamedTuple):
    tensorflow_version: str
    encoded_model_content: str
    encoded_train_X: str
    encoded_train_Y: str
    encoded_validation_X: Optional[str]
    encoded_validation_Y: Optional[str]
    user_id: str
    batch_size: int
    epochs: int
    loss_config: Union[dict, str]
    optimizer_config: Union[dict, str]
    metrics_config: Union[List[Dict], List[str]]
    fit_kwargs: Dict[str, Any]
    job_id: str
    m_d_serialized_at_ms: int
    train_X_shape: list
    train_Y_shape: list
    valid_X_shape: list
    valid_Y_shape: list
    record: bool = False


class hybridmethod:
    def __init__(self, fclass, finstance=None, doc=None):
        self.fclass = fclass
        self.finstance = finstance
        self.__doc__ = doc or fclass.__doc__
        # support use on abstract base classes
        self.__isabstractmethod__ = bool(getattr(fclass, "__isabstractmethod__", False))

    def classmethod(self, fclass):
        return type(self)(fclass, self.finstance, None)

    def instancemethod(self, finstance):
        return type(self)(self.fclass, finstance, self.__doc__)

    def __get__(self, instance, cls):
        if instance is None or self.finstance is None:
            # either bound to the class, or no instance method available
            return self.fclass.__get__(cls, None)
        return self.finstance.__get__(instance, cls)


def pickle_and_base64_decode(data: Any):
    return base64.b64encode(zlib.compress(pickle.dumps(data))).decode("utf-8")


def unpack_base64_encoded_pickle(encoded_pickle: str):
    return pickle.loads(zlib.decompress(base64.b64decode(encoded_pickle)))


def datetime_to_timestamp_ms(dt: datetime.datetime):
    return int(dt.timestamp() * 1000)


def timestamp_ms_to_datetime_str(ts: int, formatter):
    return datetime.datetime.fromtimestamp(ts / 1000).strftime(formatter)


def get_file_size_mb(file_path):
    return round(os.path.getsize(file_path) / 1024 / 1024, 4)


def get_data_size_mb(data):
    # estimated raw data size before dumping to json (json slightly compress the data)
    return round(get_deep_size(data) / 1024 / 1024, 4)


def sha256_encode(data: str):
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def get_encoded_model_content(model):
    # use random index to avoid multi thread R/W locking issue
    random_end_index = str(uuid.uuid4())
    model_name = f"model_{random_end_index}.h5"
    model.save(model_name)
    f = open(f"./{model_name}", "rb")
    model_file_content = f.read()
    f.close()
    os.remove(model_name)

    return pickle_and_base64_decode(model_file_content)


def save_tensorboard_logs(
    tensorboard_logs: Optional[str],
    not_save_log: bool,
    directory_to_save_log: Optional[str],
    job_id: Optional[str],
    f_name: str = "logs",
):
    if tensorboard_logs and not not_save_log:
        try:
            log_zip = unpack_base64_encoded_pickle(tensorboard_logs)
            path = f"{directory_to_save_log}/{f_name}.zip"
            f = open(path, "wb")
            f.write(log_zip)
            f.close()
            log_dir = f"{directory_to_save_log}/logs/{job_id}/"
            shutil.unpack_archive(path, log_dir, "zip")
            os.remove(path)
            print_receiving(f"Saved tensorboard logs file to {log_dir}")
            return path
        except Exception as e:
            return f"Failed to save tensorboard logs: {e}"


def encode_data(train_X, train_Y, validation_X, validation_Y):
    encoded_train_X = pickle_and_base64_decode(train_X) if train_X is not None else None
    encoded_train_Y = pickle_and_base64_decode(train_Y) if train_Y is not None else None
    encoded_validation_X = (
        pickle_and_base64_decode(validation_X) if validation_X is not None else None
    )
    encoded_validation_Y = (
        pickle_and_base64_decode(validation_Y) if validation_Y is not None else None
    )
    return encoded_train_X, encoded_train_Y, encoded_validation_X, encoded_validation_Y


def _load_model_from_model_content(encoded_model_content):
    model_content = unpack_base64_encoded_pickle(encoded_model_content)
    f = open("saved_model.h5", "wb")
    f.write(model_content)
    f.close()
    model = load_model("saved_model.h5")
    return model


def _deserialize_model_data(data):
    model: Model = _load_model_from_model_content(data.encoded_model_content)
    loss = (
        loss_deserialize(data.loss_config)
        if not isinstance(data.loss_config, str)
        else data.loss_config
    )
    optimizer = (
        optimizer_deserialize(data.optimizer_config)
        if not isinstance(data.optimizer_config, str)
        else data.optimizer_config
    )
    metrics = [
        metric_deserialize(metric_config)
        if not isinstance(metric_config, str)
        else metric_config
        for metric_config in data.metrics_config
    ]
    encoded_train_X = data.encoded_train_X
    encoded_train_Y = data.encoded_train_Y
    encoded_validation_X = data.encoded_validation_X
    encoded_validation_Y = data.encoded_validation_Y

    train_X = unpack_base64_encoded_pickle(encoded_train_X)
    train_Y = unpack_base64_encoded_pickle(encoded_train_Y)
    validation_X, validation_Y = None, None
    if encoded_validation_X is not None:
        validation_X = unpack_base64_encoded_pickle(encoded_validation_X)
        validation_Y = unpack_base64_encoded_pickle(encoded_validation_Y)

    result = {
        "model": model,
        "loss": loss,
        "optimizer": optimizer,
        "metrics": metrics,
        "train_X": train_X,
        "train_Y": train_Y,
        "validation_data": (validation_X, validation_Y),
    }
    return result


# Print iterations progress
def printProgressBar(
    iteration,
    total,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="█",
    printEnd="\r",
):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)
    # Print New Line on Complete
    if iteration == total:
        print()


class ProgressUpload:
    def __init__(
        self, filename, chunk_size=1250, filled_char=">", width=10, empty_char=" "
    ):
        self.filename = filename
        self.chunk_size = chunk_size
        self.file_size = os.path.getsize(filename)
        self.size_read = 0
        self.divisor = min(
            math.floor(math.log(self.file_size, 1024)) * 3, 9
        )  # cap unit at a GB
        self.unit_index = self.divisor
        self.unit = {0: "B", 3: "KiB", 6: "MiB", 9: "GiB"}[self.unit_index]
        self.divisor = 10 ** self.divisor
        self.width = width
        self.filled_char = filled_char
        self.empty_char = empty_char

    def __iter__(self):
        hide_cursor()
        progress_str = f"0 / {self.file_size / self.divisor:.2f} {self.unit}"
        progress_bar = self.empty_char * self.width
        progress_full = (
            f"\r[SENDING]: |{progress_bar}| 0.0% {progress_str} [avg: 0{self.unit}/s]"
        )
        sys.stderr.write(progress_full)
        time_start = time.time()

        with open(self.filename, "rb") as f:
            for chunk in iter(lambda: f.read(self.chunk_size), b""):
                self.size_read += len(chunk)
                # accumulate time
                time_spend = time.time() - time_start
                # calculate stuff for plotting progress bar
                rate = round(self.size_read / self.divisor / time_spend, 2)
                while not rate:
                    self.unit_index -= 3
                    self.unit = {0: "B", 3: "KiB", 6: "MiB", 9: "GiB"}[self.unit_index]
                    self.divisor /= 10 ** 3
                    rate = round(self.size_read / self.divisor / time_spend, 2)
                while rate > 1000:
                    self.unit_index += 3
                    self.unit = {0: "B", 3: "KiB", 6: "MiB", 9: "GiB"}[self.unit_index]
                    self.divisor *= 10 ** 3
                    rate = round(self.size_read / self.divisor / time_spend, 2)
                yield chunk
                percentage = self.size_read / self.file_size * 100
                completed_str = f"{self.size_read / self.divisor:.2f}"
                to_complete_str = f"{self.file_size / self.divisor:.2f} {self.unit}"
                n_prograss_bar = int(self.width * percentage / 100)
                progress_bar = (
                    n_prograss_bar * self.filled_char
                    + (self.width - n_prograss_bar) * self.empty_char
                )
                progress_str = f"{percentage:.2f} % {completed_str} / {to_complete_str}"
                progress_full = f"\r[SENDING]: |{progress_bar}| {progress_str} [avg: {rate}{self.unit}/s]"
                sys.stderr.write(progress_full)
        sys.stderr.write("\n")
        show_cursor()

    def __len__(self):
        return self.file_size


def zipdir(src, zip_path):
    ziph = zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED)
    # ziph is zipfile handle
    for root, dirs, files in os.walk(src):
        for file in files:
            ziph.write(
                os.path.join(root, file),
                os.path.relpath(os.path.join(root, file), os.path.join(src, "..")),
            )
    ziph.close()


def subprocess_to_stdout(
    cmd: List[str],
    raise_error_if_any: bool = False,
    with_stdin: bool = False,
    line_to_stop_streaming: Optional[str] = None,
    wait: bool = True,
):
    if with_stdin:
        p = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
        )
    else:
        p = subprocess.Popen(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    if wait:
        p = wait_subprocess_to_stdout(p, line_to_stop_streaming, raise_error_if_any)
    return p


def wait_subprocess_to_stdout(
    p: subprocess.Popen,
    line_to_stop_streaming: Optional[str] = None,
    raise_error_if_any: bool = False,
):
    done = False
    if p.stdout:
        for line in iter(p.stdout.readline, b""):
            decoded_str = line.rstrip().decode("utf8")
            print(decoded_str + "\r")
            if line_to_stop_streaming and line_to_stop_streaming in decoded_str:
                done = True
                break
        if not done and raise_error_if_any:
            _, _ = p.communicate()
            # If returncode != 0 then means the subprocess had an error
            if p.returncode != 0:
                raise InterruptedError()
    return p



import datetime
import os
import shutil
from time import sleep
from typing import List
from typing import Union

import pandas as pd
import plotly.graph_objs as go
import pytz
from tensorflow.keras.models import Model as TF_Model

from aibro.tools.auth import check_authentication
from aibro.tools.service.api_service import aibro_client
from aibro.tools.utils import _deserialize_model_data
from aibro.tools.utils import datetime_to_timestamp_ms
from aibro.tools.utils import hybridmethod
from aibro.tools.utils import ModelAndDataset
from aibro.tools.utils import unpack_base64_encoded_pickle
from aibro.training import _get_model_from_file
from aibro.training import Training


class Job(object):
    def __init__(self, job_id):
        self.job_id = job_id

    # @hybridmethod
    # def peek_job(cls, job_id):
    #     check_authentication()
    #     job = cls(job_id)
    #     _peek(job.job_id)

    # @peek_job.instancemethod  # type: ignore
    # def peek_job(self):
    #     check_authentication()
    #     _peek(self.job_id)

    # @hybridmethod
    # def pickup_model(
    #     cls,
    #     job_id,
    #     epochs: list = ["best_epoch"],
    #     directory=".",
    #     not_save_as_h5: bool = True,
    # ) -> Union[TF_Model, List[TF_Model], None]:
    #     check_authentication()
    #     job = cls(job_id)
    #     return _pickup_model(job.job_id, epochs, directory, not_save_as_h5)

    # @pickup_model.instancemethod  # type: ignore
    # def pickup_model(
    #     self, epochs: list = ["best_epoch"], directory=".", not_save_as_h5: bool = True
    # ) -> Union[TF_Model, List[TF_Model], None]:
    #     check_authentication()
    #     return _pickup_model(self.job_id, epochs, directory, not_save_as_h5)

    @hybridmethod
    def get_tensorboard_logs(cls, job_id, directory=".", access_token = None):
        """[reference]
        https://doc.aipaca.ai/#get_tensorboard_logs
        """
        check_authentication(access_token)
        job = cls(job_id)
        _get_tensorboard_logs(job.job_id, directory)

    @get_tensorboard_logs.instancemethod  # type: ignore
    def get_tensorboard_logs(self, directory=".", access_token = None):
        check_authentication(access_token)
        _get_tensorboard_logs(self.job_id, directory)

    @hybridmethod
    def plot_timeline(cls, job_id, char_type=None):
        """[reference]
        https://doc.aipaca.ai/#plot_timeline
        """
        job = cls(job_id)
        _plot_timeline(job.job_id, char_type)

    @plot_timeline.instancemethod  # type: ignore
    def plot_timeline(self, char_type=None):
        """[reference]
        https://doc.aipaca.ai/#plot_timeline
        """
        _plot_timeline(self.job_id, char_type)

    @hybridmethod
    def replay_job(cls, job_id, access_token = None):
        """[reference]
        https://doc.aipaca.ai/#replay_job
        """
        job = cls(job_id)
        _replay_job(job.job_id, access_token=access_token)

    @replay_job.instancemethod  # type: ignore
    def replay_job(self):
        """[reference]
        https://doc.aipaca.ai/#replay_job
        """
        _replay_job(self.job_id)

    @staticmethod
    def list_job(last_days=1, access_token = None):
        user_id = check_authentication(access_token)
        last_day_ts = datetime_to_timestamp_ms(
            datetime.datetime.now() - datetime.timedelta(days=last_days)
        )
        data = {"user_id": user_id, "last_day_ts": last_day_ts}
        resp = aibro_client.post_with_json_data("v1/list_jobs", data)

        _format_job_history(resp.json()["user_jobs"])


def _peek(job_id):
    endpoint = "status/" + job_id

    aibro_client.connect_to_socket()
    resp = aibro_client.post_with_endpoint_only(endpoint=endpoint)
    if resp.json()["job_status"] in ["COMPLETED", "CLOSING SERVER", "IDLE"]:
        # print(123)
        print("Now you can use Job.pickup_model() to pickup your model! 🎉")

        if resp.json()["job_status"] == "CANCELED":
            try:
                aibro_client.post_with_json_data("v1/stop_spot", {"job_id": job_id})
            except Exception as e:
                print(f"Cannot stop server, reason\n: {str(e)}")

    sleep(1)
    aibro_client.disconnect_socket()


def _pickup_model(
    job_id, epochs: list = ["best_epoch"], directory=".", not_save_as_h5: bool = True
) -> Union[TF_Model, List[TF_Model], None]:
    model_list = []

    for epoch in epochs:
        data = {"job_id": job_id, "epoch": epoch}
        aibro_client.connect_to_socket()
        resp = aibro_client.post_with_json_data("v1/pickup_model", data)

        if resp.status_code == 200:
            if "job_status" in resp.json().keys():
                if resp.json()["job_status"] == "CANCELED":
                    try:
                        aibro_client.post_with_json_data(
                            "v1/stop_spot", {"job_id": job_id, "error": None}
                        )
                    except Exception as e:
                        print(f"Cannot stop server, reason:\n {str(e)}")
                if resp.json()["job_status"] not in [
                    "COMPLETED",
                    "CLOSING SERVER",
                    "IDLE",
                ]:
                    sleep(1)
                    aibro_client.disconnect_socket()
                    return None
            else:
                if epoch == "best_epoch":
                    epoch = resp.json()["epoch"]
                print(f"picking up model at epoch {epoch}")
                content = resp.json()["model_ckpt_content"]
                unpacked_model_content = unpack_base64_encoded_pickle(content)

                model = _get_model_from_file(
                    model_file_content=unpacked_model_content,
                    directory_to_save_ckpt=directory,
                    not_save_as_h5=not_save_as_h5,
                    epoch=epoch,
                )
                model_list.append(model)
        else:
            aibro_client.disconnect_socket()
            print("Server error")
            return None

    print("models have been picked up successfully")
    aibro_client.disconnect_socket()
    if len(model_list) == 1:
        return model_list[0]
    return model_list


def _get_tensorboard_logs(job_id, directory="."):
    data = {"job_id": job_id}
    resp = aibro_client.post_with_json_data("v1/get_tensorboard_logs", data)
    content = resp.json()["log_zip"]
    log_zip = unpack_base64_encoded_pickle(content)
    file_name = f"{directory}/logs.zip"
    f = open(file_name, "wb")
    f.write(log_zip)
    f.close()
    log_dir = f"{directory}/logs/{job_id}/"
    shutil.unpack_archive(file_name, log_dir, "zip")
    os.remove(file_name)
    print(f"Saved logs to {log_dir}")


def _plot_timeline(job_id: str, char_type: str = None):
    assert char_type in [None, "pie"], """Available chart_type: [None, "pie"]"""
    data = {
        "job_id": job_id,
    }
    resp = aibro_client.post_with_json_data("v1/get_job_timeline", data)
    result = resp.json()
    job_created_at_ms = result["job_created_at_ms"]
    job_ended_at_ms = result["job_ended_at_ms"]
    size_m_d_trans_mb = result["size_m_d_trans_mb"]

    request_created_at_ms = result["request_created_at_ms"]
    request_fulfilled_at_ms = result["request_fulfilled_at_ms"]
    server_connected_at_ms = result["server_connected_at_ms"]
    github_key_received_at_ms = result["github_key_received_at_ms"]
    server_built_up_at_ms = result["server_built_up_at_ms"]
    env_geared_up_at_ms = result["env_geared_up_at_ms"]
    m_d_serialized_at_ms = result["m_d_serialized_at_ms"]
    m_d_received_at_ms = result["m_d_received_at_ms"]
    m_deserialized_at_ms = result["m_deserialized_at_ms"]
    training_completed_at_ms = result["training_completed_at_ms"]
    results_returned_at_ms = result["results_returned_at_ms"]
    results_received_at_ms = result["results_received_at_ms"]

    t_create_job_ms = request_created_at_ms - job_created_at_ms
    t_request_launch_ms = request_fulfilled_at_ms - request_created_at_ms
    t_server_connect_ms = server_connected_at_ms - request_fulfilled_at_ms
    t_api_trans_ms = github_key_received_at_ms - server_connected_at_ms
    t_flask_setup_ms = server_built_up_at_ms - github_key_received_at_ms
    t_gear_up_env_ms = env_geared_up_at_ms - github_key_received_at_ms
    t_m_d_serialization_ms = (
        m_d_serialized_at_ms - env_geared_up_at_ms
        if env_geared_up_at_ms
        else m_d_serialized_at_ms - job_created_at_ms
    )
    t_m_d_trans_ms = m_d_received_at_ms - m_d_serialized_at_ms
    t_m_d_deserialization_ms = m_deserialized_at_ms - m_d_received_at_ms
    t_training_ms = training_completed_at_ms - m_deserialized_at_ms
    t_m_finished_serialization_ms = results_returned_at_ms - training_completed_at_ms
    t_m_finished_trans_ms = results_received_at_ms - results_returned_at_ms
    create_string = "Job Created at: " + pd.to_datetime(
        job_created_at_ms, unit="ms"
    ).strftime("%Y/%m/%d, %H:%M:%S")
    end_string = (
        "Job Ended At: "
        + pd.to_datetime(job_ended_at_ms, unit="ms").strftime("%Y/%m/%d, %H:%M:%S")
        if job_ended_at_ms
        else ""
    )
    size_string = "Submit Model Size: " + str(size_m_d_trans_mb) + "MB"

    labels = [
        "Job Create",
        "Request Launch",
        "Instance Connect",
        "Api Transfer",
        "Server Setup",
        "M&D Serialization",
        "Env Gear Up",
        "M&D Transfer",
        "M&D Deserialization",
        "Model Training",
        "Result Serialization",
        "Result Transfer",
    ]
    values = [
        t_create_job_ms,
        t_request_launch_ms,
        t_server_connect_ms,
        t_api_trans_ms,
        t_flask_setup_ms,
        t_m_d_serialization_ms,
        t_gear_up_env_ms,
        t_m_d_trans_ms,
        t_m_d_deserialization_ms,
        t_training_ms,
        t_m_finished_serialization_ms,
        t_m_finished_trans_ms,
    ]
    colors = (len(values) * ["#494E5E"]).copy()
    colors[-3] = "#597EF9"
    y = labels
    x = [round(v / 1000, 2) if v else 0 for v in values]
    if char_type == "pie":
        y = [tmp_y + "(s)" for tmp_y in y]
        trace = [go.Pie(labels=y, values=x)]
        layout = go.Layout(
            title="Training Timeline for self.job_id: " + job_id,
            annotations=[
                go.layout.Annotation(
                    text=create_string + "<br>" + end_string + "<br>" + size_string,
                    align="right",
                    showarrow=False,
                    x=0,
                    y=0,
                    bordercolor="black",
                    borderwidth=1,
                )
            ],
        )
        fig = go.Figure(data=trace, layout=layout)
        fig.show()
    else:
        fig = go.Figure(
            go.Funnel(
                y=y,
                x=x,
                textinfo="value+percent total",
                hoverinfo="text",
                hovertemplate="Duration: %{x:.2f}s",
                insidetextfont={"color": "white"},
                marker={"color": colors},
                connector={"fillcolor": "lightblue"},
            )
        )
        fig.update_layout(
            title="Training Timeline for job_id: " + job_id,
            annotations=[
                go.layout.Annotation(
                    text=create_string + "<br>" + end_string + "<br>" + size_string,
                    align="left",
                    showarrow=False,
                    x=40,
                    y=0,
                    bordercolor="black",
                    borderwidth=1,
                )
            ],
        )
        fig.show()


def _format_job_history(jobs):
    print(
        f"{'Date (EST)' : <20} {'Job id' : <40} {'Status' : <12} {'Substatus' : <12} {'Current epoch' : <20} Description"  # noqa: E501
    )
    for job in jobs:
        dt = datetime.datetime.fromtimestamp(job["job_created_ms"] // 1000)
        formatted_date = dt.astimezone(pytz.timezone("US/Eastern")).strftime(
            "%m/%d/%Y %H:%M:%S"
        )
        current_epoch = job["current_epoch"]
        status = job["status"]
        substatus = job["substatus"]
        job_id = job["job_id"]
        description = job["description"]
        print(
            f"{f'{formatted_date}' : <20} {f'{job_id}' : <40} {f'{status}' : <12} {f'{substatus}' : <12} {f'{current_epoch}' : <20} {description}"  # noqa
        )


def _replay_job(
    job_id: str,
    description: str = "",
    directory_to_save_ckpt: str = ".",
    directory_to_save_log: str = ".",
    wait_request_s: int = 10000,
    access_token=None,
):
    user_id = check_authentication(access_token)
    post_data = {"job_id": job_id, "user_id": user_id}
    print(f"[REPLAYING]: grabing history model & data of job {job_id}.")
    resp = aibro_client.post_with_json_data("v1/replay_job", post_data)
    job = resp.json()["job"]
    model_and_data = resp.json()["model_and_data"]
    data = ModelAndDataset(**model_and_data)
    data_decode = _deserialize_model_data(data)
    print("[REPLAYING]: cloning a new job to start replay.")
    description = description if description != "" else job["description"]
    job_id, result_model, history = Training.online_fit(  # type: ignore
        model=data_decode["model"],
        train_X=data_decode["train_X"],
        train_Y=data_decode["train_Y"],
        validation_data=data_decode["validation_data"],
        machine_ids=job["instance_type"],
        batch_size=data.batch_size,
        epochs=data.epochs,
        description=description,
        directory_to_save_ckpt=directory_to_save_ckpt,
        directory_to_save_log=directory_to_save_log,
        wait_request_s=wait_request_s,
    )

    assert history
    print(f"history.history: {history.history}")
    print(f"history.params: {history.params}")
    print(f"OnlinePickupModel: {result_model}")
